from .utils_name import LOG_NAME_AGG

from .aggregation_stages import AggregationStage
from .agreggation_operations import AggregationOperation
from .aggregation import AggregationExecutor

from . import aggregation_stages as agg_st
from . import agg_op_regex as agg_op_re
from . import agg_op_comparator as agg_op_comp
from . import agg_op_operators as agg_op_op
from . import agg_op_logical as agg_op_logical

from . import aggregation_decorators as agg_de

__all__ = ['LOG_NAME_AGG', 'AggregationStage',
           'AggregationOperation', 'AggregationExecutor',
           'agg_st', 'agg_op_re',
           'agg_op_comp', 'agg_op_op',
           'agg_op_logical', 'agg_de']




from abc import ABC

from .aggregation_functions import \
    AggregationFunctionValue, AggregationFunctionList
from .agreggation_operations import AggregationOperation


class _AggregationOperationProjectComparison(AggregationOperation,
                                             AggregationFunctionValue, ABC):
    def __init__(self, first_value, second_value):
        super().__init__([first_value, second_value])


class AggOpProjectEq(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$eq'


class AggOpProjectNe(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$ne'


class AggOpProjectIn(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$in'


class AggOpProjectNin(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$nin'


class AggOpProjectGt(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$gt'


class AggOpProjectGte(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$gte'


class AggOpProjectLt(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$lt'


class AggOpProjectLte(_AggregationOperationProjectComparison):

    @classmethod
    def key_word(cls) -> str:
        return '$lte'


class AggOpMatchEq(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$eq'


class AggOpMatchNe(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$ne'


class AggOpMatchIn(AggregationFunctionList, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$in'


class AggOpMatchNin(AggregationFunctionList, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$nin'


class AggOpMatchGt(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$gt'


class AggOpMatchGte(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$gte'


class AggOpMatchLt(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$lt'


class AggOpMatchLte(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$lte'



from abc import ABC

from .aggregation_functions import AggregationFunction, AggregationFunctionDict, \
    AggregationFunctionValue


class AggregationStage(AggregationFunction, ABC):
    pass


class AggStMatch(AggregationFunctionDict, AggregationStage):

    @classmethod
    def key_word(cls) -> str:
        return '$match'


class AggStGroup(AggregationFunctionDict, AggregationStage):

    def __init__(self, id_mapping: dict = None, mapping: dict = None):
        super().__init__()
        if id_mapping is None:
            id_mapping = {}
        id_mapping = {self.ID: id_mapping}
        if mapping is None:
            mapping = {}
        self.data = {self.key_word(): {**id_mapping, **mapping}}

    def add_id(self, mapping: dict):
        self[self.key_word()][self.ID] = mapping

    @classmethod
    def key_word(cls) -> str:
        return '$group'


class AggStProject(AggregationFunctionDict, AggregationStage):

    def __init__(self, mapping: dict = None, id_present: bool = None):
        super().__init__()
        if mapping is None:
            mapping = {}
        if id_present is not None:
            mapping[self.ID] = 1 if id_present else 0
        self.data = {self.key_word(): mapping}

    def presence_attr(self, att: str, presence: bool):
        self[att] = 1 if presence else 0

    @classmethod
    def key_word(cls) -> str:
        return '$project'


class AggStFacet(AggregationFunctionDict, AggregationStage):

    @classmethod
    def key_word(cls) -> str:
        return '$facet'


class AggStUnwind(AggregationFunctionValue, AggregationStage):

    @classmethod
    def key_word(cls) -> str:
        return '$unwind'

    def __init__(self, unwind_value: str):
        super().__init__(unwind_value)


class AggStSort(AggregationFunctionDict, AggregationStage):

    def __init__(self, mapping: dict = None):
        super().__init__()
        if mapping is None:
            mapping = {}
        self.data = {self.key_word(): mapping}

    @classmethod
    def key_word(cls) -> str:
        return '$sort'

    def add_fields(self, sorted_dict: dict):
        for x, y in sorted_dict.items():
            self[x] = y

    def add_field(self, field: str, asc: bool):
        self[field] = 1 if asc else -1


class AggStOut(AggregationFunctionValue, AggregationStage):

    @classmethod
    def key_word(cls) -> str:
        return '$out'

    def __init__(self, collection_out: str):
        super().__init__(collection_out)


class AggStLimit(AggregationFunctionValue, AggregationStage):

    @classmethod
    def key_word(cls) -> str:
        return '$limit'

    def __init__(self, number_limit: int):
        super().__init__(number_limit)


class AggStSkip(AggregationFunctionValue, AggregationStage):

    @classmethod
    def key_word(cls) -> str:
        return '$skip'

    def __init__(self, number_skip: int):
        super().__init__(number_skip)


class AggStCount(AggregationFunctionValue, AggregationStage):

    def __init__(self, var_count: str):
        super().__init__(var_count)

    @classmethod
    def key_word(cls) -> str:
        return '$count'



import functools
import logging
from typing import List, Union

import pymongo
from pymongo.cursor import Cursor

from . import LOG_NAME_AGG, AggregationExecutor


def aggregation_decorator(function):
    @functools.wraps(function)
    def wrapper(self, *args,
                connection: pymongo.collection = None, **kwargs) -> List[dict]:
        exe = self._generate_aggregation_executor()\
            if connection is None else AggregationExecutor(connection)
        agg = function(self, exe, *args, **kwargs)
        r = agg.execute()
        return r if r is None else list(r)

    return wrapper


def aggregation_decorator_cursor(function):
    @functools.wraps(function)
    def wrapper(self, *args,
                connection: pymongo.collection = None, **kwargs) -> Union[Cursor, None]:
        exe = self._generate_aggregation_executor()\
            if connection is None else AggregationExecutor(connection)
        agg = function(self, exe, *args, **kwargs)
        return agg.execute()

    return wrapper


def aggregation_decorator_debug(function):
    @functools.wraps(function)
    def wrapper(self, *args,
                connection: pymongo.collection = None, **kwargs) -> List[dict]:
        logging.basicConfig(level=logging.DEBUG)
        log = logging.getLogger(LOG_NAME_AGG)
        exe = self._generate_aggregation_executor()\
            if connection is None else AggregationExecutor(connection)
        agg = function(self, exe, *args, **kwargs)
        log.debug('\n' + repr(agg))
        r = agg.execute()
        return r if r is None else list(r)

    return wrapper


def aggregation_decorator_alone(function):
    def wrapper(connection: pymongo.collection, *args, **kwargs) -> List[dict]:
        agg = function(AggregationExecutor(connection), *args, **kwargs)
        r = agg.execute()
        return r if r is None else list(r)

    return wrapper


def aggregation_decorator_cursor_alone(function):
    def wrapper(connection: pymongo.collection, *args, **kwargs) -> Union[Cursor, None]:
        agg = function(AggregationExecutor(connection), *args, **kwargs)
        return agg.execute()

    return wrapper


def aggregation_decorator_debug_alone(function):
    def wrapper(connection: pymongo.collection, *args, **kwargs) -> List[dict]:
        logging.basicConfig(level=logging.DEBUG)
        log = logging.getLogger(LOG_NAME_AGG)
        agg = function(AggregationExecutor(connection), *args, **kwargs)
        log.debug('\n' + repr(agg))
        r = agg.execute()
        return r if r is None else list(r)

    return wrapper



from __future__ import annotations

import json
from abc import ABC

from bson import ObjectId
from pymongo.cursor import Cursor


class ObjetoMongoAbstract(ABC):

    def __init__(self, _id=None, **kwargs):
        self._id = _id
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self) -> ObjectId:
        return self._id

    def get_dict(self, id_mongo=True, id_as_string=False) -> dict:
        """
        Return all variables from this object, similar to vars(object)
         with exception for _id
        @param id_mongo: True -> Return dict with _id
        @param id_as_string: True -> Return _id as str if id_mongo is True
        @return: dict with object variables, similar to vars(object)
        """
        d = vars(self).copy()
        if not id_mongo:
            d.pop('_id')
        elif id_as_string and self._id is not None:
            d['_id'] = str(self._id)
        for x, y in self.get_attr_nested_objects().items():
            obj = d.get(x, None)
            if obj is None:
                continue
            if isinstance(obj, list):
                d[x] = y.generar_list_dicts_from_list_objects(d[x], False, False)
            else:
                d[x] = d[x].get_dict(False, False)
        return d

    def get_dict_no_id(self) -> dict:
        """
        Facade for get_dict(id_mongo=False). Return all variables from this object,
         similar to vars(object)
         with exception for _id
        @return: dict with object variables, similar to vars(object)
        """
        return self.get_dict(id_mongo=False)

    def serialize(self, id_mongo=True) -> str:
        """
        Serialize object to JSON format
        @param id_mongo: True -> Append _id to json
        @return: str in Json format
        """
        return json.dumps(self.get_dict(id_mongo=id_mongo, id_as_string=True))

    @staticmethod
    def serialize_all(objetos, id_mongo=True) -> str:
        """
        Serialize all objects to JSON format
        @param objetos: List for object to convert to JSON
        @param id_mongo: True -> Append _id to json
        @return: str in Json format
        """
        return json.dumps(
            ObjetoMongoAbstract.generar_list_dicts_from_list_objects(objetos,
                                                                     id_mongo=id_mongo,
                                                                     id_as_string=True))

    @staticmethod
    def prepare_dict_for_generated_object(dictionary: dict, attr: dict) -> dict:
        for x, y in attr.items():
            obj = dictionary.get(x, None)
            if obj is None:
                continue
            if isinstance(obj, list):
                dictionary[x] = y.generar_objects_from_list_dicts(obj)
            else:
                dictionary[x] = y.generar_object_from_dict(obj)
        return dictionary

    @classmethod
    def generar_object_from_dict(cls, dictionary):
        if dictionary is None:
            return None
        return cls(**cls.prepare_dict_for_generated_object(
            dictionary, cls.get_attr_nested_objects()))

    def __repr__(self):
        return str(self.get_dict(True, True))

    def __str__(self):
        return f'Class: {self.__class__.__name__} -> ObjectID {self._id}'

    @classmethod
    def generar_objects_from_list_dicts(cls, dictionaries: list | Cursor):
        return [cls.generar_object_from_dict(dictionary) for dictionary in dictionaries]

    @staticmethod
    def generar_list_dicts_from_list_objects(lista_objetos: list,
                                             id_mongo=True, id_as_string=False):
        return [c.get_dict(id_mongo=id_mongo, id_as_string=id_as_string)
                for c in lista_objetos]

    @staticmethod
    def get_attr_nested_objects() -> dict:
        """
        Attributes which are objects of the class ObjetoMongoAbstract,
         all of them. Override parents methods.
        @return: dict format {name_attr: class_attr}
        """
        return {}



import re

from .aggregation_functions import AggregationFunctionValue
from .agreggation_operations import AggregationOperation


class AggOpRegex(AggregationOperation):

    def __init__(self, regex: str, options: str = ''):
        super().__init__()
        self.data = {self.key_word(): regex,
                     '$options': options}

    @staticmethod
    def get_options_regex(insensitive: bool = False,
                          multiline_anchor: bool = False,
                          ignore_space_regex: bool = False,
                          allow_dot_character: bool = False):
        s = []
        if multiline_anchor:
            s.append('m')
        if ignore_space_regex:
            s.append('x')
        if allow_dot_character:
            s.append('s')
        if insensitive:
            s.append('i')
        return ''.join(s)

    @classmethod
    def insensitive_search(cls, search: str):
        return cls(regex='^' + search + '$', options='i')

    @classmethod
    def key_word(cls) -> str:
        return '$regex'


class AggOpRegexRe(AggregationFunctionValue, AggregationOperation):

    def __init__(self, regex: re.Pattern):
        super().__init__(regex)
        self.logger.debug('\nLa query no sera ejecutable en MongoDB sin usar pymongo.\n'
                          'Query is not going to be able to be executed in MongoDB'
                          ' without using pymongo.')

    @classmethod
    def key_word(cls) -> str:
        return '$regex'



import _collections_abc
from typing import List, Union

import pymongo.collection
from pymongo.cursor import Cursor

from . import AggregationStage


def _pretty_query(query, total_tab='\t') -> str:
    st = ''
    _st = '\n' + total_tab
    if issubclass(type(query), _collections_abc.MutableMapping):
        st = _st.join([
            "'{x}': {y}".format(x=x,
                                y=_pretty_query(query=y, total_tab=total_tab + '\t'))
            for x, y in query.items()])
        st = '{\n' + total_tab + st + '\n' + total_tab[:-1] + '}'
    elif issubclass(type(query), list):
        st = st + _st.join(_pretty_query(x, total_tab + '\t') for x in query)
    elif issubclass(type(query), str):
        st = f"'{query}'"
    else:
        st = f"{str(query)}"
    return st


class AggregationExecutor:
    def __init__(self, collection_to_execute: pymongo.collection.Collection):
        self.__query: List[Union[AggregationStage, dict]] = []
        self.__collection = collection_to_execute

    def add_step(self, a: Union[AggregationStage, dict]):
        self.__query.append(a)

    def add_steps(self, *a: Union[AggregationStage, dict]):
        self.__query.extend(a)

    def add_steps_list(self, steps=List[Union[AggregationStage, dict]]):
        self.__query.extend(steps)

    def execute(self) -> Union[Cursor, None]:
        if len(self.__query) == 0:
            return None
        return self.__collection.aggregate(self.__query)

    def _see_query(self, sep=',\n\t') -> str:
        return sep.join([str(a) for a in self.__query])

    def pretty(self) -> str:
        aux = _pretty_query(self.__query)
        query = '[\n\t' + aux + '\n]'
        return f'db.{self.__collection.name}.aggregate({query})'

    def __str__(self):
        query = '[\n\t' + self._see_query() + '\n]'
        return f'db.{self.__collection.name}.aggregate({query})'

    def __repr__(self):
        query = '[' + self._see_query(",") + ']'
        return f'db.{self.__collection.name}.aggregate({query})'




import logging
from abc import ABC, abstractmethod
from collections import UserDict

from . import LOG_NAME_AGG
from ..exceptions import MongoManagerAggregationException


class AggregationFunction(UserDict, ABC):
    ID = '_id'

    @staticmethod
    def _logger() -> logging.Logger:
        return logging.getLogger(LOG_NAME_AGG)

    @property
    def logger(self) -> logging.Logger:
        return self._logger()

    @classmethod
    @abstractmethod
    def key_word(cls) -> str:
        pass


class AggregationFunctionDict(AggregationFunction, ABC):
    def __init__(self, mapping: dict = None):
        super().__init__()
        if mapping is None:
            mapping = {}
        self.data = {self.key_word(): mapping}

    def __setitem__(self, key, value):
        self[self.key_word()][key] = value


class AggregationFunctionList(AggregationFunction, ABC):
    def __init__(self, mapping: list = None):
        super().__init__()
        if mapping is None:
            mapping = []
        self.data = {self.key_word(): mapping}

    def add_item(self, item):
        self['data'][self.key_word()].append(item)

    def __setitem__(self, key, value):
        raise MongoManagerAggregationException('No se ha definido ninguna funcion'
                                               ' para añadir valores a este paso'
                                               ' de la query.')


class AggregationFunctionValue(AggregationFunction, ABC):

    def __init__(self, mapping):
        super().__init__()
        self.data = {self.key_word(): mapping}

    def __setitem__(self, key, value):
        raise MongoManagerAggregationException('No se ha definido ninguna funcion'
                                               ' para añadir valores a'
                                               ' este paso de la query.')



from .exceptions import MongoManagerException, \
    MongoManagerAggregationException
from .mongo_manager import MongoManager
from .entity import ObjetoMongoAbstract
from .repository import RepositoryBase

MONGO_MANAGER_ASCENDING = 1
"""Ascending sort order."""
MONGO_MANAGER_DESCENDING = -1
"""Descending sort order."""

ObjectMongoAbstract = ObjetoMongoAbstract

__all__ = ['MongoManagerException', 'MongoManagerAggregationException',
           'MongoManager',
           'ObjetoMongoAbstract', 'RepositoryBase',
           'MONGO_MANAGER_ASCENDING', 'MONGO_MANAGER_DESCENDING']



from abc import ABC

from .aggregation_functions import AggregationFunction


class AggregationOperation(AggregationFunction, ABC):

    def get_var_aggregation(self, name_var: str) -> dict:
        return {name_var: self}

    @classmethod
    def gen_operation_aggregations_name(cls, name_var, *args, **kwargs):
        return {name_var: cls(*args, **kwargs)}





from abc import ABC
from typing import List

from .agreggation_operations import AggregationOperation
from .aggregation_functions import AggregationFunctionValue, AggregationFunctionList, \
    AggregationFunctionDict


class _AggregationOperationString(AggregationFunctionValue,
                                  AggregationOperation, ABC):

    def __init__(self, expression: str):
        super().__init__(expression)


class AggOpToLower(_AggregationOperationString):

    @classmethod
    def key_word(cls) -> str:
        return '$toLower'


class AggOpToUpper(_AggregationOperationString):

    @classmethod
    def key_word(cls) -> str:
        return '$toUpper'


class AggOpConcat(AggregationFunctionList, AggregationOperation):

    def __init__(self, expression: List[str]):
        super().__init__(expression)

    @classmethod
    def key_word(cls) -> str:
        return '$concat'


class AggOpSubstr(AggregationFunctionValue, AggregationOperation):

    def __init__(self, expression: str, start: int, end: int):
        super().__init__([expression, start, end])

    @classmethod
    def key_word(cls) -> str:
        return '$substr'


class AggOpAvg(AggregationFunctionValue, AggregationOperation):
    def __init__(self, expression: str):
        super().__init__(expression)

    @classmethod
    def key_word(cls) -> str:
        return '$avg'


class AggOpSum(AggregationFunctionValue, AggregationOperation):
    @classmethod
    def key_word(cls) -> str:
        return '$sum'


class AggOpRound(AggregationFunctionValue, AggregationOperation):

    def __init__(self, expression: str, decimal: int = 0):
        super().__init__([expression, decimal])

    @classmethod
    def key_word(cls) -> str:
        return '$round'


class AggOpTrunc(AggregationFunctionValue, AggregationOperation):

    def __init__(self, expression: str, decimal: int = 0):
        super().__init__([expression, decimal])

    @classmethod
    def key_word(cls) -> str:
        return '$trunc'


class AggOpPush(AggregationFunctionDict, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$push'


class AggOpConcatArrays(AggregationFunctionList, AggregationOperation):
    @classmethod
    def key_word(cls) -> str:
        return '$concatArrays'



from __future__ import annotations
from pymongo import MongoClient

from .patrones import SingletonMeta

_mongo_manager_gl = None


class MongoManager(metaclass=SingletonMeta):
    __bd = None

    def __init__(self, username: str = '', password: str = '',
                 db: str = '', auth_source: str = '',
                 bd_online: bool = False, port_local: int = 27017,
                 url_online='', authenticated=True) -> None:
        """
        Crea la instancia conectada a la collecion en cuestion.
        """
        if bd_online:
            url = url_online
        else:
            if authenticated:
                url = f'mongodb://{username}:{password}@localhost:{port_local}'
            else:
                url = f'mongodb://localhost:{port_local}'
        self.__client = MongoClient(url, authSource=auth_source) \
            if authenticated and auth_source != '' else MongoClient(url)
        self.__bd = self.__client[db]
        global _mongo_manager_gl
        _mongo_manager_gl = self

    @property
    def bd(self):
        return self.__bd

    def collection(self, collection):
        return self.bd[collection]

    @classmethod
    def _destroy(cls):
        del cls._instances[cls]



# Aggregation Pipelines Stages

def aggregate_match(diccionario: dict) -> dict:
    return {'$match': diccionario}


def aggregate_group(diccionario: dict) -> dict:
    return {'$group': diccionario}


def aggregate_project(diccionario: dict) -> dict:
    return {'$project': diccionario}


def aggregate_unwind(valor: str) -> dict:
    return {'$unwind': valor}


def aggregate_sort(diccionario: dict) -> dict:
    return {'$sort': diccionario}


def aggregate_out(collection_name) -> dict:
    return {'$out':  collection_name}


def aggregate_limit(limit: int) -> dict:
    return {'$limit': limit}


def aggregate_skip(skip: int) -> dict:
    return {'$skip': skip}


def aggregate_facet(diccionario: dict) -> dict:
    return {'$facet': diccionario}


def aggregate_count(nombre_counter: str) -> dict:
    return {'$count': nombre_counter}


# Query Operators

#   Comparison

def mongo_eq(valor) -> dict:
    return {'$eq': valor}


def mongo_ne(valor) -> dict:
    return {'$ne': valor}


def mongo_in(valor: list) -> dict:
    return {'$in': valor}


def mongo_nin(valor: list) -> dict:
    return {'$nin': valor}


def mongo_gt(valor: int) -> dict:
    return {'$gt': valor}


def mongo_gte(valor: int) -> dict:
    return {'$gte': valor}


def mongo_lt(valor: list) -> dict:
    return {'$lt': valor}


def mongo_lte(valor: list) -> dict:
    return {'$lte': valor}


#   Logical

def mongo_and(valor: list) -> dict:
    return {'$and': valor}


def mongo_or(valor: list) -> dict:
    return {'$or': valor}


def mongo_nor(valor: list) -> dict:
    return {'$nor': valor}


def mongo_not(condicion_dict: dict) -> dict:
    return {'$not': condicion_dict}


#   Others

def mongo_exists(existe: bool = True) -> dict:
    return {'$exists': existe}


def mongo_regex(regex, options) -> dict:
    return {'$regex': regex, '$options': options}


def mongo_regex_compare_insensitive(valor: str) -> dict:
    return mongo_regex('^' + valor + '$', 'i')


# Fields Operators

#   String Operators

def mongo_lower(valor: str) -> dict:
    return {'$toLower': valor}


def mongo_upper(valor: str) -> dict:
    return {'$toUpper': valor}


def mongo_concat(valor: list) -> dict:
    return {'$concat': valor}


#   Number Operators


def mongo_avg(valor: str) -> dict:
    return {'$avg': valor}


def mongo_sum(valor: int) -> dict:
    return {'$sum': valor}


def mongo_sum_field(field: str) -> dict:
    return {'$sum': field}


def mongo_round(field: str, decimal: int) -> dict:
    return {'$round': [field, decimal]}


def mongo_trunc(field: str, decimal: int) -> dict:
    return {'$trunc': [field, decimal]}


#   Array Operators

def mongo_push(valor: dict) -> dict:
    return {'$push': valor}


def mongo_concat_arrays(arrays: list) -> dict:
    return {'$concatArrays': arrays}



import setuptools
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mongo-manager-juan-palma-borda",
    version="0.9.7",
    author='Juan Palma Borda',
    author_email='juanpalmaborda@hotmail.com',
    description='Libreria para manejar objetos almacenados en MongoDB, '
                'usando la referencia de los CRUDRepository de SpringBoot',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/muerterauda/mongo_manager",
    project_urls={
        "Bug Tracker": "https://github.com/muerterauda/mongo_manager/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7",
    install_requires=[
        'pymongo>=3.12.1'
    ],
)



from abc import ABC

from .aggregation_functions import AggregationFunctionList, AggregationFunctionValue
from .agreggation_operations import AggregationOperation


class _AggregationOperationLogical(AggregationFunctionList, AggregationOperation, ABC):
    pass


class AggOpAnd(_AggregationOperationLogical):

    @classmethod
    def key_word(cls) -> str:
        return '$and'


class AggOpOr(_AggregationOperationLogical):

    @classmethod
    def key_word(cls) -> str:
        return '$and'


class AggOpNor(_AggregationOperationLogical):

    @classmethod
    def key_word(cls) -> str:
        return '$and'


class AggOpProjectNot(_AggregationOperationLogical):

    @classmethod
    def key_word(cls) -> str:
        return '$not'


class AggOpMatchNot(AggregationFunctionValue, AggregationOperation, ABC):

    @classmethod
    def key_word(cls) -> str:
        return '$not'


class AggOpExists(AggregationFunctionValue, AggregationOperation):

    @classmethod
    def key_word(cls) -> str:
        return '$exists'

    def __init__(self, exists: bool):
        super().__init__(exists)



from __future__ import annotations

from collections.abc import Iterable
from typing import TypeVar, Type, Generic, Optional

import pymongo
from pymongo.results import UpdateResult, InsertManyResult, \
    InsertOneResult, DeleteResult

from .. import ObjetoMongoAbstract, MongoManagerException
from ..patrones import SingletonMeta
from ..aggregations.aggregation import AggregationExecutor
from ..mongo_utils import aggregate_out, aggregate_match
from bson import ObjectId

T_O = TypeVar('T_O', bound=ObjetoMongoAbstract)


class RepositoryBase(Generic[T_O], metaclass=SingletonMeta):

    def __init__(self, collection: str, clase: Type[T_O],
                 connection_collection=None) -> None:
        __metaclass__ = SingletonMeta  # noqa: F841
        if connection_collection is None:
            from ..mongo_manager import _mongo_manager_gl as mongo_manager_gl
            if mongo_manager_gl is None:
                raise MongoManagerException(
                    '\nNo se ha inicializado la base de datos, '
                    'crear con MongoManager(), '
                    'ni se ha dado una conexion al repositorio.'
                    '\nMongoManager is not initialized, '
                    'create it with MongoManager(), '
                    'or provide a different connection to the repository.')
            connection_collection = mongo_manager_gl.collection(collection)
        self.__collection = connection_collection
        self.__clase = clase

    @property
    def collection(self) -> pymongo.collection.Collection:
        """
        @return: Pymongo collection | Coleccion Pymongo
        """
        return self.__collection

    @property
    def clase(self):
        """
        @return: Class used in the repository to
         convert data o objects| Clase del repositorio
        """
        return self.__clase

    def _generate_aggregation_executor(self) -> AggregationExecutor:
        return AggregationExecutor(collection_to_execute=self.collection)

    def count_many(self, filter_dict: dict = None) -> int:
        if filter_dict is None:
            filter_dict = {}
        return self.collection.count_documents(filter_dict)

    def count_all(self) -> int:
        return self.collection.count_documents({})

    def estimated_count(self) -> int:
        return self.collection.estimated_document_count()

    def find_one(self, filter_dict: dict = None) -> T_O:
        if filter_dict is None:
            filter_dict = {}
        return self.clase.generar_object_from_dict(self.collection.find_one(
            filter_dict))

    def find_many(self, filter_dict: dict = None,
                  skip: int = 0, limit: int = 1000, sort=None) -> list[T_O]:
        if filter_dict is None:
            filter_dict = {}
        object_list = self.collection.find(filter_dict).skip(skip).limit(limit if limit != -1 else 0)
        if sort is not None:
            object_list.sort(sort)
        return self.clase.generar_objects_from_list_dicts(object_list)

    def find_all(self, skip: int = 0, limit: int = 1000, sort=None) -> list[T_O]:
        return self.find_many(skip=skip, limit=limit, sort=sort)

    def find_by_id(self, id_mongo) -> T_O:
        return self.clase.generar_object_from_dict(
            self.collection.find_one({'_id': ObjectId(id_mongo)}))

    def delete_object(self, objeto: T_O) -> DeleteResult:
        if objeto.id is not None:
            return self.delete_by_id(objeto.id)

    def delete_by_id(self, id_mongo) -> DeleteResult:
        return self.collection.delete_one({'_id': ObjectId(id_mongo)})

    def insert_one(self, objeto: T_O, id_mongo=False) -> InsertOneResult:
        return self.collection.insert_one(objeto.get_dict_no_id()
                                          if not id_mongo else objeto.get_dict())

    def insert_many(self, object_list: list[T_O], id_mongo=False) -> InsertManyResult | None:
        if (object_list is None or
                not isinstance(object_list, Iterable) or len(object_list) == 0):
            return None
        return self.collection.insert_many(
            self.clase.generar_list_dicts_from_list_objects(
                object_list, id_mongo=id_mongo))

    def insert_one_raw(self, object_dict: dict):
        return self.insert_one(self.clase.generar_object_from_dict(object_dict))

    def insert_or_replace_id(self, objeto: T_O, upsert: bool = True):
        return self._replace_by_id(objeto.id, objeto, upsert)

    def replace_by_id(self, id_mongo, objeto: T_O) -> UpdateResult:
        return self._replace_by_id(id_mongo, objeto, False)

    def _replace_by_id(self, id_mongo, objeto: T_O, upsert: bool = False) -> UpdateResult:
        return self.collection.replace_one({"_id": id_mongo}, objeto.get_dict(), upsert)

    def replace_many_by_id(self, object_list: list[T_O], upsert: bool = False):
        for x in object_list:
            self._replace_by_id(x.id, x, upsert)

    def update_by_id(self, id_mongo, objeto_dict: dict) -> UpdateResult:
        return self.collection.update_one({"_id": id_mongo}, {"$set": objeto_dict})

    def update_many(self, filter_dict: dict = None,
                    objeto_dict: dict = None) -> Optional[UpdateResult]:
        if objeto_dict is None:
            return None
        if filter_dict is None:
            filter_dict = {}
        self.collection.update_many(filter_dict, {"$set": objeto_dict})

    def copy_collection_db(self, name_new_collection: str,
                           copy_match_condition: dict = None):
        out = [aggregate_out(name_new_collection)]
        if copy_match_condition is not None:
            match = aggregate_match(copy_match_condition)
            out = [match] + out
        return self.collection.aggregate(out)

    def drop_collection(self, *, drop=False):
        if not drop:
            raise MongoManagerException('Medida de seguridad, fallo al '
                                        'intentar eliminar la coleccion\n'
                                        'Security fail, try to drop '
                                        'entire collection avoided.')
        self.collection.delete_many({})
